/**
 * libSpiff - XSPF playlist handling library
 *
 * Copyright (C) 2006 Sebastian Pipping
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Sebastian Pipping, webmaster@hartwork.org
 */

#include "../Source/Spiff/Spiff.h"
#include <stdio.h>
using namespace Spiff;



class DemoPlaylist : public SpiffReaderCallback {

private:
	void open(XML_Char * type) {
		PORT_PRINTF(_PT("%s\n"), type);
	}

	void close() {
		PORT_PRINTF(_PT("\n"));
	}

	void printDeleteKeyValue(const XML_Char * key, XML_Char * value) {
		if (value == NULL) {
			PORT_PRINTF(_PT("\t%s: -\n"), key);
		} else {
			PORT_PRINTF(_PT("\t%s: '%s'\n"), key, value);
			delete [] value;
		}
	}

	void printKeyValue(const XML_Char * key, int value) {
		if (value == -1) {
			PORT_PRINTF(_PT("\t%s: -\n"), key);
		} else {
			PORT_PRINTF(_PT("\t%s: '%i'\n"), key, value);
		}
	}

	void printDeleteKeyValuesString(const XML_Char * key, SpiffStringList * values) {
		if (values == NULL) {
			PORT_PRINTF(_PT("\t%s: 0 entries\n"), key);
		} else {
			const int size = static_cast<int>(values->size());
			PORT_PRINTF(_PT("\t%s: %i entr%s\n"), key, size, (size == 1) ? _PT("y") : _PT("ies"));
			SpiffStringList::iterator iter = values->begin();
			while (iter != values->end()) {
				delete [] *iter;
				iter++;
			}
			delete values;
		}
	}

	void printDeleteKeyValuesMetaLink(const XML_Char * key, SpiffMetaLinkList * values) {
		if (values == NULL) {
			PORT_PRINTF(_PT("\t%s: 0 entries\n"), key);
		} else {
			const int size = static_cast<int>(values->size());
			PORT_PRINTF(_PT("\t%s: %i entr%s\n"), key, size, (size == 1) ? _PT("y") : _PT("ies"));
			SpiffMetaLinkList::iterator iter = values->begin();
			while (iter != values->end()) {
				SpiffMetaLink & entry = *iter;
				delete [] entry.first;
				delete [] entry.second;
				iter++;
			}
			delete values;
		}
	}

	void printDeleteKeyValuesAttribution(const XML_Char * key, SpiffAttributionList * values) {
		if (values == NULL) {
			PORT_PRINTF(_PT("\t%s: 0 entries (0 identifiers, 0 locations)\n"), key);
		} else {
			const int size = static_cast<int>(values->size());
			SpiffAttributionList::iterator iter = values->begin();
			int identifiers = 0;
			int locations = 0;
			while (iter != values->end()) {
				SpiffAttribution & entry = *iter;
				if (entry.first) {
					identifiers++;
				} else {
					locations++;
				}
				delete [] entry.second;
				iter++;
			}
			delete values;
			PORT_PRINTF(_PT("\t%s: %i entr%s (%i identifier%s, %i location%s)\n"), key,
					size, (size == 1) ? _PT("y") : _PT("ies"), identifiers,
					(identifiers == 1) ? _PT("") : _PT("s"), locations,
					(locations == 1) ? _PT("") : _PT("s"));
		}
	}
	void printDeleteKeyValue(const XML_Char * key, SpiffDateTime * value) {
		if (value == NULL) {
			PORT_PRINTF(_PT("\t%s: -\n"), key);
		} else {
			PORT_PRINTF(_PT("\t%s: %04i-%02i-%02i %02i:%02i:%02i %s%02i:%02i\n"), key,
					value->year, value->month, value->day, value->hour, value->minutes,
					value->seconds, (value->distHours < 0) ? _PT("-") : _PT("+"),
					std::abs(value->distHours), std::abs(value->distMinutes));
			delete value;
		}
	}

	void addTrack(SpiffTrack & track) {
		open(_PT("Track"));
			printDeleteKeyValue(_PT("Album"), track.stealAlbum());
			printDeleteKeyValue(_PT("Annotation"), track.stealAnnotation());
			printDeleteKeyValue(_PT("Creator"), track.stealCreator());
			printKeyValue(_PT("Duration"), track.getDuration());
			printDeleteKeyValuesString(_PT("Identifiers"), track.stealIdentifier());
			printDeleteKeyValue(_PT("Image"), track.stealImage());
			printDeleteKeyValue(_PT("Info"), track.stealInfo());
			printDeleteKeyValuesMetaLink(_PT("Links"), track.stealLink());
			printDeleteKeyValuesString(_PT("Locations"), track.stealLocation());
			printDeleteKeyValuesMetaLink(_PT("Metas"), track.stealMeta());
			printDeleteKeyValue(_PT("Title"), track.stealTitle());
			printKeyValue(_PT("TrackNum"), track.getTrackNum());
		close();
	}

	void setProperties(SpiffProps & props) {
		open(_PT("Props"));
			printDeleteKeyValue(_PT("Annotation"), props.stealAnnotation());
			printDeleteKeyValuesAttribution(_PT("Attribution"), props.stealAttribution());
			printDeleteKeyValue(_PT("Creator"), props.stealCreator());
			printDeleteKeyValue(_PT("Date"), props.stealDate());
			printDeleteKeyValue(_PT("Identifier"), props.stealIdentifier());
			printDeleteKeyValue(_PT("Image"), props.stealImage());
			printDeleteKeyValue(_PT("Info"), props.stealInfo());
			printDeleteKeyValue(_PT("License"), props.stealLicense());
			printDeleteKeyValuesMetaLink(_PT("Links"), props.stealLink());
			printDeleteKeyValue(_PT("Location"), props.stealLocation());
			printDeleteKeyValuesMetaLink(_PT("Metas"), props.stealMeta());
			printDeleteKeyValue(_PT("Title"), props.stealTitle());
			printKeyValue(_PT("Version"), props.getVersion());
		close();
	}

};



int PORT_MAIN() {
	SpiffReader reader;
	const int res = reader.parse(_PT("Playlist.xspf"), new DemoPlaylist());
	if (res != SPIFF_PARSER_SUCCESS) {
		PORT_PRINTF(_PT("Error %i at line %i: '%s'\n"), res, reader.getErrorLine(), reader.getErrorText());
	} else {
		PORT_PRINTF(_PT("Everything fine.\n"));
	}
	return 0;
}
