/**
 * libSpiff - XSPF playlist handling library
 *
 * Copyright (C) 2006 Sebastian Pipping
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Sebastian Pipping, webmaster@hartwork.org
 */

#include "SpiffIndentPropsWriter.h"
#include "SpiffProps.h"

namespace Spiff {



SpiffIndentPropsWriter::SpiffIndentPropsWriter(SpiffProps * props) :
		output(NULL), props(props), trackCount(-1), version(-1) {

}



SpiffIndentPropsWriter::~SpiffIndentPropsWriter() {

}



void SpiffIndentPropsWriter::writeAnnotation() {
	const XML_Char * const annotation = this->props->getAnnotation();
	if (annotation != NULL) {
		(*this->output) << _PT("\t<annotation>") << annotation << _PT("</annotation>\n");
	}
}



void SpiffIndentPropsWriter::writeAttribution() {
	const SpiffAttributionList * const attributions = this->props->getAttribution();
	if (attributions != NULL) {
		SpiffAttributionList::const_iterator iter = attributions->begin();
		if (iter != attributions->end()) {
			(*this->output) << _PT("\t<attribution>\n");
			do {
				const SpiffAttribution & entry = *iter;
				if (entry.first) {
					(*this->output) << _PT("\t\t<identifier>") << entry.second
							<< _PT("</identifier>\n");
				} else {
					(*this->output) << _PT("\t\t<location>") << entry.second
							<< _PT("</location>\n");
				}
				iter++;
			} while (iter != attributions->end());
			(*this->output) << _PT("\t</attribution>\n");
		}
	}
}



void SpiffIndentPropsWriter::writeCreator() {
	const XML_Char * const creator = this->props->getCreator();
	if (creator != NULL) {
		(*this->output) << _PT("\t<creator>") << creator << _PT("</creator>\n");
	}
}



void SpiffIndentPropsWriter::writeDate() {
	const SpiffDateTime * const dateTime = this->props->getDate();
	if (dateTime != NULL) {
		XML_Char buffer[10 + 1 + 8 + 6 + 1];
		::PORT_SPRINTF(buffer, _PT("%04i-%02i-%02iT%02i:%02i:%02i%s%02i:%02i"), 
				dateTime->year, dateTime->month, dateTime->day, dateTime->hour,
				dateTime->minutes, dateTime->seconds,
				(dateTime->distHours < 0) ? _PT("-") : _PT("+"),
				std::abs(dateTime->distHours), std::abs(dateTime->distMinutes));
		(*this->output) << _PT("\t<date>") << buffer << _PT("</date>\n");
	}
}



void SpiffIndentPropsWriter::writeExtensions() {
	// Nothing to do
}



void SpiffIndentPropsWriter::writeIdentifier() {
	const XML_Char * const identifier = this->props->getIdentifier();
	if (identifier != NULL) {
		(*this->output) << _PT("\t<identifier>") << identifier << _PT("</identifier>\n");
	}
}



void SpiffIndentPropsWriter::writeImage() {
	const XML_Char * const image = this->props->getImage();
	if (image != NULL) {
		(*this->output) << _PT("\t<image>") << image << _PT("</image>\n");
	}
}



void SpiffIndentPropsWriter::writeInfo() {
	const XML_Char * const info = this->props->getInfo();
	if (info != NULL) {
		(*this->output) << _PT("\t<info>") << info << _PT("</info>\n");
	}
}



void SpiffIndentPropsWriter::writeLicense() {
	const XML_Char * const license = this->props->getLicense();
	if (license != NULL) {
		(*this->output) << _PT("\t<license>") << license << _PT("</license>\n");
	}
}



void SpiffIndentPropsWriter::writeLink() {
	const SpiffMetaLinkList * const links = this->props->getLink();
	if (links != NULL) {
		SpiffMetaLinkList::const_iterator iter = links->begin();
		while (iter != links->end()) {
			const SpiffMetaLink & entry = *iter;
			(*this->output) << _PT("\t<link rel=\"") << entry.first << _PT("\">")
					<< entry.second << _PT("</link>\n");
			iter++;
		}
	}
}



void SpiffIndentPropsWriter::writeLocation() {
	const XML_Char * const location = this->props->getLocation();
	if (location != NULL) {
		(*this->output) << _PT("\t<location>") << location << _PT("</location>\n");
	}
}



void SpiffIndentPropsWriter::writeMeta() {
	const SpiffMetaLinkList * const metas = this->props->getMeta();
	if (metas != NULL) {
		SpiffMetaLinkList::const_iterator iter = metas->begin();
		while (iter != metas->end()) {
			const SpiffMetaLink & entry = *iter;
			(*this->output) << _PT("\t<meta rel=\"") << entry.first << _PT("\">")
					<< entry.second << _PT("</meta>\n");
			iter++;
		}
	}
}



void SpiffIndentPropsWriter::writePlaylistClose() {
	(*this->output) << _PT("</playlist>\n");
}



void SpiffIndentPropsWriter::writePlaylistOpen() {
	(*this->output) << _PT("<?xml version=\"1.0\" encoding=\"utf-8\"?>\n");
	(*this->output) << _PT("<playlist version=\"") << this->version
			<< _PT("\" xmlns=\"http://xspf.org/ns/0/\">\n");
}



void SpiffIndentPropsWriter::writeTitle() {
	const XML_Char * const title = this->props->getTitle();
	if (title != NULL) {
		(*this->output) << _PT("\t<title>") << title << _PT("</title>\n");
	}
}



void SpiffIndentPropsWriter::writeTrackListClose() {
	if ((this->trackCount > 0)) {
		(*this->output) << _PT("\t</trackList>\n");
	}
}



void SpiffIndentPropsWriter::writeTrackListOpen() {
	if ((this->trackCount < 1)) {
		if (this->version > 0) {
			// Empty trackList allowed
			(*this->output) << _PT("\t<trackList />\n");
		} else {
			// Empty trackList forbidden
			(*this->output) << _PT("\t<trackList>\n\t\t<track />\n\t</trackList>\n"); 
		}
	} else {
		(*this->output) << _PT("\t<trackList>\n");
	}
}



void SpiffIndentPropsWriter::init(basic_ostringstream<XML_Char> & output, int version, int trackCount) {
	this->output = &output;
	this->trackCount = trackCount;

	if ((version < 0) || (version > 1)) {
		this->version = 1;
	} else {
		this->version = version;
	}
}



void SpiffIndentPropsWriter::writeHead() {
	writePlaylistOpen();
	writeTitle();
	writeCreator();
	writeAnnotation();
	writeInfo();
	writeLocation();
	writeIdentifier();
	writeImage();
	writeDate();
	writeLicense();
	writeAttribution();
	writeLink();
	writeMeta();
	writeExtensions();
	writeTrackListOpen();
}


void SpiffIndentPropsWriter::writeFoot() {
	writeTrackListClose();
	writePlaylistClose();
}



}
