/**
 * libSpiff - XSPF playlist handling library
 *
 * Copyright (C) 2006 Sebastian Pipping
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Sebastian Pipping, webmaster@hartwork.org
 */

#include "SpiffReader.h"
#include "SpiffProps.h"
#include "SpiffTrack.h"
#include "SpiffReaderCallback.h"
#include "SpiffToolbox.h"
#include <stdio.h>
#include <algorithm>
#include <stdlib.h> // atoi

namespace Spiff {



SpiffReader::SpiffReader() : props(NULL), track(NULL), callback(callback),
		firstPlaylistAnnotation(true), firstPlaylistAttribution(true),
		firstPlaylistCreator(true),	firstPlaylistDate(true), firstPlaylistIdentifier(true),
        firstPlaylistImage(true), firstPlaylistInfo(true), firstPlaylistLicense(true),
		firstPlaylistLocation(true), firstPlaylistTitle(true), firstPlaylistTrackList(true),
		firstTrackTitle(true), firstTrackCreator(true),	firstTrackAnnotation(true),
		firstTrackInfo(true), firstTrackImage(true), firstTrackAlbum(true),
		firstTrackTrackNum(true), firstTrackDuration(true), insideExtension(false),
		errorCode(SPIFF_PARSER_SUCCESS), errorLine(-1), version(-1), firstTrack(true) {

}


		
void SpiffReader::makeReusable() {
	// Reset everything but the error state
	this->stack.clear();
	if (this->props != NULL) {
		delete this->props;
		this->props = NULL;
	}
	if (this->track != NULL) {
		delete this->track;
		this->track = NULL;
	}
	this->callback = NULL;
	this->accum.clear();
	this->lastRelValue.clear();

	this->firstPlaylistAnnotation = true;
	this->firstPlaylistAttribution = true;
	this->firstPlaylistCreator = true;
	this->firstPlaylistDate = true;
	this->firstPlaylistIdentifier = true;
	this->firstPlaylistImage = true;
	this->firstPlaylistInfo = true;
	this->firstPlaylistLicense = true;
	this->firstPlaylistLocation = true;
	this->firstPlaylistTitle = true;
	this->firstPlaylistTrackList = true;

	this->firstTrackTitle = true;
	this->firstTrackCreator = true;
	this->firstTrackAnnotation = true;
	this->firstTrackInfo = true;
	this->firstTrackImage = true;
	this->firstTrackAlbum = true;
	this->firstTrackTrackNum = true;
	this->firstTrackDuration = true;

	this->firstTrack = true;

	this->insideExtension = false;
	this->version = -1;
}



SpiffReader::~SpiffReader() {
	if (this->props != NULL) {
		delete this->props;
		this->props = NULL;
	}
	if (this->track != NULL) {
		delete this->track;
		this->track = NULL;
	}
}



int SpiffReader::parse(const XML_Char * filename, SpiffReaderCallback * callback) {
	clearError();

	// Check filename
	if (filename == NULL) {
		setError(SPIFF_PARSER_ERROR_NO_INPUT, _PT("Filename must not be NULL."));
		return this->errorCode;
	}

	// Set callback, NULL is no problem
	this->callback = callback;

	// Create parser
	this->parser = XML_ParserCreate(NULL);

	// Put class pointer into user data
	XML_SetUserData(this->parser, this);

	// Register handlers
	XML_SetElementHandler(this->parser, masterStart, masterEnd);
	XML_SetCharacterDataHandler(this->parser, masterCharacters);

	// Open file
	FILE * file = ::PORT_FOPEN(filename, _PT("r"));
	if (file == NULL) {
		setError(SPIFF_PARSER_ERROR_NO_INPUT, _PT("File '%s' could not be read."), filename);
		return this->errorCode;
	}

	// Get filesize
	::fseek(file, 0, SEEK_END);
	const long filesize = ::ftell(file);
	::fseek(file, 0, SEEK_SET);

	// Read and parse file
	void * buffer;
	if (filesize > SPIFF_MAX_BLOCK_SIZE) {
		// In several blocks
		long sizeLeft = filesize;
		while (sizeLeft > 0) {
			const long blockSize = std::min<long>(sizeLeft, SPIFF_MAX_BLOCK_SIZE);
			buffer = malloc(blockSize);
			::fread(buffer, 1, blockSize, file);
			sizeLeft -= blockSize;
			if (XML_Parse(this->parser, reinterpret_cast<char *>(buffer), blockSize, sizeLeft == 0) == XML_STATUS_ERROR) {
				if (this->errorCode == SPIFF_PARSER_SUCCESS) {
					const XML_Error expatCode = XML_GetErrorCode(this->parser);
					setError(SPIFF_PARSER_ERROR_EXPAT + static_cast<int>(expatCode),
							_PT("Expat error '%s'"),
							XML_ErrorString(expatCode));
				}
				break;
			}
		}
		::fclose(file);
	} else {
		// One single go
		buffer = XML_GetBuffer(this->parser, filesize);
		::fread(buffer, 1, filesize, file);
		::fclose(file);

		if (XML_ParseBuffer(this->parser, filesize, 1) == XML_STATUS_ERROR) {
			if (this->errorCode == SPIFF_PARSER_SUCCESS) {
				const XML_Error expatCode = XML_GetErrorCode(this->parser);
				setError(SPIFF_PARSER_ERROR_EXPAT + static_cast<int>(expatCode),
						_PT("Expat error '%s'"),
						XML_ErrorString(expatCode));
			}
		}
	}

	// Cleanup
	XML_ParserFree(this->parser);

	makeReusable();
	return this->errorCode;
}



bool SpiffReader::handleStartOne(const XML_Char * name, const XML_Char ** atts) {
	if (::PORT_STRCMP(name, _PT("playlist"))) {
		setError(SPIFF_PARSER_ERROR_ELEMENT_FORBIDDEN,
				_PT("Element '%s' not allowed."),
				name);
		return false;
	}

	if (!handlePlaylistAttribs(atts)) {
		return false;
	}

	this->stack.push(TAG_PLAYLIST);
	this->props = newProps();
	this->props->setVersion(this->version);
	return true;
}



bool SpiffReader::handleStartTwo(const XML_Char * name, const XML_Char ** atts) {
	switch (name[0]) {
	case _PT('a'):
		switch (name[1]) {
		case _PT('n'):
			if (::PORT_STRCMP(name + 2, _PT("notation"))) {
				break;
			}
			if (!this->firstPlaylistAnnotation) {
				setError(SPIFF_PARSER_ERROR_ELEMENT_TOOMANY,
						_PT("Only one 'annotation' allowed for 'playlist'."));
				return false;
			} else if (!handleNoAttribs(atts)) {
				return false;
			} else {
				this->firstPlaylistAnnotation = false;
				this->stack.push(TAG_PLAYLIST_ANNOTATION);
				return true;
			}
			break;

		case _PT('t'):
			if (::PORT_STRCMP(name + 2, _PT("tribution"))) {
				break;
			}
			if (!this->firstPlaylistAttribution) {
				setError(SPIFF_PARSER_ERROR_ELEMENT_TOOMANY,
						_PT("Only one 'attribution' allowed for 'playlist'."));
				return false;
			} else if (!handleNoAttribs(atts)) {
				return false;
			} else {
				this->firstPlaylistAttribution = false;
				this->stack.push(TAG_PLAYLIST_ATTRIBUTION);
				return true;
			}
			break;

		}
		break;

	case _PT('c'):
		if (::PORT_STRCMP(name + 1, _PT("reator"))) {
			break;
		}
		if (!this->firstPlaylistCreator) {
			setError(SPIFF_PARSER_ERROR_ELEMENT_TOOMANY,
					_PT("Only one 'creator' allowed for 'playlist'."));
			return false;
		} else if (!handleNoAttribs(atts)) {
			return false;
		} else {
			this->firstPlaylistCreator = false;
			this->stack.push(TAG_PLAYLIST_CREATOR);
			return true;
		}
		break;

	case _PT('d'):
		if (::PORT_STRCMP(name + 1, _PT("ate"))) {
			break;
		}
		if (!this->firstPlaylistDate) {
			setError(SPIFF_PARSER_ERROR_ELEMENT_TOOMANY,
					_PT("Only one 'date' allowed for 'playlist'."));
			return false;
		} else if (!handleNoAttribs(atts)) {
			return false;
		} else {
			this->firstPlaylistDate = false;
			this->stack.push(TAG_PLAYLIST_DATE);
			return true;
		}
		break;

	case _PT('e'):
		if (::PORT_STRCMP(name + 1, _PT("xtension"))) {
			break;
		}

		// Tag only allowed in v1
		if (this->version == 0) {
			setError(SPIFF_PARSER_ERROR_ELEMENT_FORBIDDEN,
					_PT("Element '%s' not allowed."),
					name);
			return false;
		}

		if (handleExtensionAttribs(atts)) {
			this->insideExtension = true;
			return handleExtensionStart(name, atts);
		} else {
			return false;
		}
		break;

	case _PT('i'):
		switch (name[1]) {
		case _PT('d'):
			if (::PORT_STRCMP(name + 2, _PT("entifier"))) {
				break;
			}
			if (!this->firstPlaylistIdentifier) {
				setError(SPIFF_PARSER_ERROR_ELEMENT_TOOMANY,
						_PT("Only one 'identifier' allowed for 'playlist'."));
				return false;
			} else if (!handleNoAttribs(atts)) {
				return false;
			} else {
				this->firstPlaylistIdentifier = false;
				this->stack.push(TAG_PLAYLIST_IDENTIFIER);
				return true;
			}
			break;

		case _PT('m'):
			if (::PORT_STRCMP(name + 2, _PT("age"))) {
				break;
			}
			if (!this->firstPlaylistImage) {
				setError(SPIFF_PARSER_ERROR_ELEMENT_TOOMANY,
						_PT("Only one 'image' allowed for 'playlist'."));
				return false;
			} else if (!handleNoAttribs(atts)) {
				return false;
			} else {
				this->firstPlaylistImage = false;
				this->stack.push(TAG_PLAYLIST_IMAGE);
				return true;
			}
			break;

		case _PT('n'):
			if (::PORT_STRCMP(name + 2, _PT("fo"))) {
				break;
			}
			if (!this->firstPlaylistInfo) {
				setError(SPIFF_PARSER_ERROR_ELEMENT_TOOMANY,
						_PT("Only one 'info' allowed for 'playlist'."));
				return false;
			} else if (!handleNoAttribs(atts)) {
				return false;
			} else {
				this->firstPlaylistInfo = false;
				this->stack.push(TAG_PLAYLIST_INFO);
				return true;
			}
			break;

		}
		break;

	case _PT('l'):
		switch (name[1]) {
		case _PT('i'):
			switch (name[2]) {
			case _PT('c'):
				if (::PORT_STRCMP(name + 3, _PT("ense"))) {
					break;
				}
				if (!this->firstPlaylistLicense) {
					setError(SPIFF_PARSER_ERROR_ELEMENT_TOOMANY,
							_PT("Only one 'license' allowed for 'playlist'."));
					return false;
				} else if (!handleNoAttribs(atts)) {
					return false;
				} else {
					this->firstPlaylistLicense = false;
					this->stack.push(TAG_PLAYLIST_LICENSE);
					return true;
				}
				break;

			case _PT('n'):
				if (::PORT_STRCMP(name + 3, _PT("k"))) {
					break;
				}
				if (handleMetaLinkAttribs(atts)) {
					this->stack.push(TAG_PLAYLIST_LINK);
					this->lastRelValue.assign(atts[1]);
					return true;
				} else {
					return false;
				}
				break;

			}
			break;

		case _PT('o'):
			if (::PORT_STRCMP(name + 2, _PT("cation"))) {
				break;
			}
			if (!this->firstPlaylistLocation) {
				setError(SPIFF_PARSER_ERROR_ELEMENT_TOOMANY,
						_PT("Only one 'location' allowed for 'playlist'."));
				return false;
			} else if (!handleNoAttribs(atts)) {
				return false;
			} else {
				this->firstPlaylistLocation = false;
				this->stack.push(TAG_PLAYLIST_LOCATION);
				return true;
			}
			break;

		}
		break;

	case _PT('m'):
		if (::PORT_STRCMP(name + 1, _PT("eta"))) {
			break;
		}
		if (handleMetaLinkAttribs(atts)) {
			this->stack.push(TAG_PLAYLIST_META);
			this->lastRelValue.assign(atts[1]);
			return true;
		} else {
			return false;
		}
		break;

	case _PT('t'):
		switch (name[1]) {
		case _PT('i'):
			if (::PORT_STRCMP(name + 2, _PT("tle"))) {
				break;
			}
			if (!this->firstPlaylistTitle) {
				setError(SPIFF_PARSER_ERROR_ELEMENT_TOOMANY,
						_PT("Only one 'title' allowed for 'playlist'."));
				return false;
			} else if (!handleNoAttribs(atts)) {
				return false;
			} else {
				this->firstPlaylistTitle = false;
				this->stack.push(TAG_PLAYLIST_TITLE);
				return true;
			}
			break;

		case _PT('r'):
			if (::PORT_STRCMP(name + 2, _PT("ackList"))) {
				break;
			}
			if (!this->firstPlaylistTrackList) {
				setError(SPIFF_PARSER_ERROR_ELEMENT_TOOMANY,
						_PT("Only one 'trackList' allowed for 'playlist'."));
				return false;
			} else if (!handleNoAttribs(atts)) {
				return false;
			} else {
				this->firstPlaylistTrackList = false;
				this->stack.push(TAG_PLAYLIST_TRACKLIST);
				return true;
			}
			break;

		}
		break;

	}

	setError(SPIFF_PARSER_ERROR_ELEMENT_FORBIDDEN,
			_PT("Element '%s' not allowed."),
			name);
	return false;
}



bool SpiffReader::handleStartThree(const XML_Char * name, const XML_Char ** atts) {
	switch (this->stack.top()) {
	case TAG_PLAYLIST_ATTRIBUTION:
		switch (name[0]) {
		case _PT('i'):
			if (::PORT_STRCMP(name + 1, _PT("dentifier"))) {
				break;
			}
			if (!handleNoAttribs(atts)) {
				return false;
			}
			this->stack.push(TAG_PLAYLIST_ATTRIBUTION_IDENTIFIER);
			return true;

		case _PT('l'):
			if (::PORT_STRCMP(name + 1, _PT("ocation"))) {
				break;
			}
			if (!handleNoAttribs(atts)) {
				return false;
			}
			this->stack.push(TAG_PLAYLIST_ATTRIBUTION_LOCATION);
			return true;

		}
		break;

	case TAG_PLAYLIST_TRACKLIST:
		if (::PORT_STRCMP(name, _PT("track"))) {
			break;
		}
		if (!handleNoAttribs(atts)) {
			return false;
		}
		this->firstTrack = false;
		this->stack.push(TAG_PLAYLIST_TRACKLIST_TRACK);
		this->track = newTrack();
		return true;

	}

	setError(SPIFF_PARSER_ERROR_ELEMENT_FORBIDDEN,
			_PT("Element '%s' not allowed."),
			name);
	return false;
}



bool SpiffReader::handleStartFour(const XML_Char * name, const XML_Char ** atts) {
	if (this->stack.top() != TAG_PLAYLIST_TRACKLIST_TRACK) {
		return false;
	}

	switch (name[0]) {
	case _PT('a'):
		switch (name[1]) {
		case _PT('l'):
			if (::PORT_STRCMP(name + 2, _PT("bum"))) {
				break;
			}
			if (!this->firstTrackAlbum) {
				setError(SPIFF_PARSER_ERROR_ELEMENT_TOOMANY,
						_PT("Only one 'album' allowed for 'track'."));
				return false;
			} else if (!handleNoAttribs(atts)) {
				return false;
			} else {
				this->firstTrackAlbum = false;
				this->stack.push(TAG_PLAYLIST_TRACKLIST_TRACK_ALBUM);
				return true;
			}
			break;

		case _PT('n'):
			if (::PORT_STRCMP(name + 2, _PT("notation"))) {
				break;
			}
			if (!this->firstTrackAnnotation) {
				setError(SPIFF_PARSER_ERROR_ELEMENT_TOOMANY,
						_PT("Only one 'annotation' allowed for 'track'."));
				return false;
			} else if (!handleNoAttribs(atts)) {
				return false;
			} else {
				this->firstTrackAnnotation = false;
				this->stack.push(TAG_PLAYLIST_TRACKLIST_TRACK_ANNOTATION);
				return true;
			}
			break;

		}
		break;

	case _PT('c'):
		if (::PORT_STRCMP(name + 1, _PT("reator"))) {
			break;
		}
		if (!this->firstTrackCreator) {
			setError(SPIFF_PARSER_ERROR_ELEMENT_TOOMANY,
					_PT("Only one 'creator' allowed for 'track'."));
			return false;
		} else if (!handleNoAttribs(atts)) {
			return false;
		} else {
			this->firstTrackCreator = false;
			this->stack.push(TAG_PLAYLIST_TRACKLIST_TRACK_CREATOR);
			return true;
		}
		break;

	case _PT('d'):
		if (::PORT_STRCMP(name + 1, _PT("uration"))) {
			break;
		}
		if (!this->firstTrackDuration) {
			setError(SPIFF_PARSER_ERROR_ELEMENT_TOOMANY,
					_PT("Only one 'duration' allowed for 'track'."));
			return false;
		} else if (!handleNoAttribs(atts)) {
			return false;
		} else {
			this->firstTrackDuration = false;
			this->stack.push(TAG_PLAYLIST_TRACKLIST_TRACK_DURATION);
			return true;
		}
		break;

	case _PT('e'):
		if (::PORT_STRCMP(name + 1, _PT("xtension"))) {
			break;
		}

		// Tag only allowed in v1
		if (this->version == 0) {
			setError(SPIFF_PARSER_ERROR_ELEMENT_FORBIDDEN,
					_PT("Element '%s' not allowed."),
					name);
			return false;
		}

		if (handleExtensionAttribs(atts)) {
			this->insideExtension = true;
			return handleExtensionStart(name, atts);
		} else {
			return false;
		}
		break;

	case _PT('i'):
		switch (name[1]) {
		case _PT('d'):
			if (::PORT_STRCMP(name + 2, _PT("entifier"))) {
				break;
			}
			if (!handleNoAttribs(atts)) {
				return false;
			}
			this->stack.push(TAG_PLAYLIST_TRACKLIST_TRACK_IDENTIFIER);
			return true;

		case _PT('m'):
			if (::PORT_STRCMP(name + 2, _PT("age"))) {
				break;
			}
			if (!this->firstTrackImage) {
				setError(SPIFF_PARSER_ERROR_ELEMENT_TOOMANY,
						_PT("Only one 'image' allowed for 'track'."));
				return false;
			} else if (!handleNoAttribs(atts)) {
				return false;
			} else {
				this->firstTrackImage = false;
				this->stack.push(TAG_PLAYLIST_TRACKLIST_TRACK_IMAGE);
				return true;
			}
			break;

		case _PT('n'):
			if (::PORT_STRCMP(name + 2, _PT("fo"))) {
				break;
			}
			if (!this->firstTrackInfo) {
				setError(SPIFF_PARSER_ERROR_ELEMENT_TOOMANY,
						_PT("Only one 'info' allowed for 'track'."));
				return false;
			} else if (!handleNoAttribs(atts)) {
				return false;
			} else {
				this->firstTrackInfo = false;
				this->stack.push(TAG_PLAYLIST_TRACKLIST_TRACK_INFO);
				return true;
			}
			break;

		}
		break;

	case _PT('l'):
		switch (name[1]) {
		case _PT('i'):
			if (::PORT_STRCMP(name + 2, _PT("nk"))) {
				break;
			}
			if (handleMetaLinkAttribs(atts)) {
				this->stack.push(TAG_PLAYLIST_TRACKLIST_TRACK_LINK);
				this->lastRelValue.assign(atts[1]);
				return true;
			} else {
				return false;
			}
			break;

		case _PT('o'):
			if (::PORT_STRCMP(name + 2, _PT("cation"))) {
				break;
			}
			if (!handleNoAttribs(atts)) {
				return false;
			}
			this->stack.push(TAG_PLAYLIST_TRACKLIST_TRACK_LOCATION);
			return true;

		}
		break;

	case _PT('m'):
		if (::PORT_STRCMP(name + 1, _PT("eta"))) {
			break;
		}
		if (handleMetaLinkAttribs(atts)) {
			this->stack.push(TAG_PLAYLIST_TRACKLIST_TRACK_META);
			this->lastRelValue.assign(atts[1]);
			return true;
		} else {
			return false;
		}
		break;

	case _PT('t'):
		switch (name[1]) {
		case _PT('i'):
			if (::PORT_STRCMP(name + 2, _PT("tle"))) {
				break;
			}
			if (!this->firstTrackTitle) {
				setError(SPIFF_PARSER_ERROR_ELEMENT_TOOMANY,
						_PT("Only one 'title' allowed for 'track'."));
				return false;
			} else if (!handleNoAttribs(atts)) {
				return false;
			} else {
				this->firstTrackTitle = false;
				this->stack.push(TAG_PLAYLIST_TRACKLIST_TRACK_TITLE);
				return true;
			}
			break;

		case _PT('r'):
			if (::PORT_STRCMP(name + 2, _PT("ackNum"))) {
				break;
			}
			if (!this->firstTrackTrackNum) {
				setError(SPIFF_PARSER_ERROR_ELEMENT_TOOMANY,
						_PT("Only one 'trackNum' allowed for 'track'."));
				return false;
			} else if (!handleNoAttribs(atts)) {
				return false;
			} else {
				this->firstTrackTrackNum = false;
				this->stack.push(TAG_PLAYLIST_TRACKLIST_TRACK_TRACKNUM);
				return true;
			}
			break;

		}
		break;

	}

	setError(SPIFF_PARSER_ERROR_ELEMENT_FORBIDDEN,
			_PT("Element '%s' not allowed."),
			name);
	return false;
}



bool SpiffReader::handleExtensionStart(const XML_Char * name, const XML_Char ** atts) {
	switch (this->stack.getSize() + 1) {
	case 2:
		// We only get called on this level for playlist.extension
		// Name and attributes have already been checked.
		this->stack.push(TAG_PLAYLIST_EXTENSION);
		return true;;

	case 4:
		if (this->stack.top() == TAG_PLAYLIST_TRACKLIST_TRACK) {
			// Name and attributes have already been checked.
			this->stack.push(TAG_PLAYLIST_TRACKLIST_TRACK_EXTENSION);
			return true;
		}
		break;

	}

	this->stack.push(TAG_UNKNOWN);
	return true;
}



bool SpiffReader::handleExtensionEnd(const XML_Char * name) {
	this->stack.pop();
	return true;
}



bool SpiffReader::handleExtensionCharacters(const XML_Char * s, int len) {
	return true;
}



void SpiffReader::handleStart(const XML_Char * name, const XML_Char ** atts) {
	bool res;
	if (this->insideExtension) {
		res = handleExtensionStart(name, atts);
	} else {
		switch (this->stack.getSize() + 1) {
		case 1:
			res = handleStartOne(name, atts);
			break;

		case 2:
			res = handleStartTwo(name, atts);
			break;

		case 3:
			res = handleStartThree(name, atts);
			break;

		case 4:
			res = handleStartFour(name, atts);
			break;

		case 5:
			setError(SPIFF_PARSER_ERROR_ELEMENT_FORBIDDEN,
					_PT("Element '%s' not allowed."),
					name);
			res = false;
			break;

		}
	}

	if (!res) {
		stop();
	}
}



bool SpiffReader::handleEndOne(const XML_Char * name) {
	if (this->firstPlaylistTrackList) {
		setError(SPIFF_PARSER_ERROR_ELEMENT_MISSING,
				_PT("Element 'trackList' missing."));
		return false;
	}

	// Call property callback
	if (this->callback != NULL) {
		this->callback->setProperties(*this->props);
	}
	delete this->props;
	this->props = NULL;
	return true;
}



bool SpiffReader::handleEndTwo(const XML_Char * name) {
	const XML_Char * const finalAccum = this->accum.c_str();

	switch (this->stack.top()) {
	case TAG_PLAYLIST_ANNOTATION:
		this->props->setAnnotation(finalAccum);
		break;
/*
	case TAG_PLAYLIST_ATTRIBUTION:
		break;
*/
	case TAG_PLAYLIST_CREATOR:
		this->props->setCreator(finalAccum);
		break;

	case TAG_PLAYLIST_DATE:
		SpiffDateTime dateTime;
		if (!extractDateTime(finalAccum, &dateTime)) {
			setError(SPIFF_PARSER_ERROR_CONTENT_INVALID,
					_PT("Content of 'date' is not a dateTime."));
			return false;
		} else {
			this->props->setDate(&dateTime);
		}
		break;
/*
	case TAG_PLAYLIST_EXTENSION:
		break;
*/
	case TAG_PLAYLIST_IDENTIFIER:
		if (!isURI(finalAccum)) {
			setError(SPIFF_PARSER_ERROR_CONTENT_INVALID,
					_PT("Content of 'identifier' is not a URI."));
			return false;
		} else {
			this->props->setIdentifier(finalAccum);
		}
		break;

	case TAG_PLAYLIST_IMAGE:
		if (!isURI(finalAccum)) {
			setError(SPIFF_PARSER_ERROR_CONTENT_INVALID,
					_PT("Content of 'image' is not a URI."));
			return false;
		} else {
			this->props->setImage(finalAccum);
		}
		break;

	case TAG_PLAYLIST_INFO:
		if (!isURI(finalAccum)) {
			setError(SPIFF_PARSER_ERROR_CONTENT_INVALID,
					_PT("Content of 'info' is not a URI."));
			return false;
		} else {
			this->props->setInfo(finalAccum);
		}
		break;

	case TAG_PLAYLIST_LICENSE:
		if (!isURI(finalAccum)) {
			setError(SPIFF_PARSER_ERROR_CONTENT_INVALID,
					_PT("Content of 'license' is not a URI."));
			return false;
		} else {
			this->props->setLicense(finalAccum);
		}
		break;

	case TAG_PLAYLIST_LINK:
		if (!isURI(finalAccum)) {
			setError(SPIFF_PARSER_ERROR_CONTENT_INVALID,
					_PT("Content of 'link' is not a URI."));
			return false;
		} else {
			this->props->addLink(this->lastRelValue.c_str(), finalAccum);
		}
		break;

	case TAG_PLAYLIST_LOCATION:
		if (!isURI(finalAccum)) {
			setError(SPIFF_PARSER_ERROR_CONTENT_INVALID,
					_PT("Content of 'location' is not a URI."));
			return false;
		} else {
			this->props->setLocation(finalAccum);
		}
		break;

	case TAG_PLAYLIST_META:
		this->props->addMeta(this->lastRelValue.c_str(), finalAccum);
		break;

	case TAG_PLAYLIST_TITLE:
		this->props->setTitle(finalAccum);
		break;

	case TAG_PLAYLIST_TRACKLIST:
		// Check if empty for v0
		if ((this->version == 0) && (this->firstTrack)) {
			setError(SPIFF_PARSER_ERROR_ELEMENT_MISSING,
					_PT("Element 'track' missing."));
			return false;
		}
		break;

	}

	this->accum.clear();
	return true;
}



bool SpiffReader::handleEndThree(const XML_Char * name) {
	const XML_Char * const finalAccum = this->accum.c_str();

	switch (this->stack.top()) {
	case TAG_PLAYLIST_ATTRIBUTION_IDENTIFIER:
		if (!isURI(finalAccum)) {
			setError(SPIFF_PARSER_ERROR_CONTENT_INVALID,
					_PT("Content of 'identifier' is not a URI."));
			return false;
		} else {
			this->props->addAttributionIdentifier(finalAccum);
		}
		break;

	case TAG_PLAYLIST_ATTRIBUTION_LOCATION:
		if (!isURI(finalAccum)) {
			setError(SPIFF_PARSER_ERROR_CONTENT_INVALID,
					_PT("Content of 'location' is not a URI."));
			return false;
		} else {
			this->props->addAttributionLocation(finalAccum);
		}
		break;

	case TAG_PLAYLIST_TRACKLIST_TRACK:
		// Call track callback
		if (this->callback != NULL) {
			this->callback->addTrack(*this->track);
		}
		delete this->track;
		this->track = NULL;

		this->firstTrackTitle = true;
		this->firstTrackCreator = true;
		this->firstTrackAnnotation = true;
		this->firstTrackInfo = true;
		this->firstTrackImage = true;
		this->firstTrackAlbum = true;
		this->firstTrackTrackNum = true;
		this->firstTrackDuration = true;
	}

	this->accum.clear();
	return true;
}



bool SpiffReader::handleEndFour(const XML_Char * name) {
	const XML_Char * const finalAccum = this->accum.c_str();

	switch (this->stack.top()) {
	case TAG_PLAYLIST_TRACKLIST_TRACK_ALBUM:
		this->track->setAlbum(finalAccum);
		break;

	case TAG_PLAYLIST_TRACKLIST_TRACK_ANNOTATION:
		this->track->setAnnotation(finalAccum);
		break;

	case TAG_PLAYLIST_TRACKLIST_TRACK_CREATOR:
		this->track->setCreator(finalAccum);
		break;

	case TAG_PLAYLIST_TRACKLIST_TRACK_DURATION:
		int duration;
		if (!extractInteger(finalAccum, 0, &duration)) {
			setError(SPIFF_PARSER_ERROR_CONTENT_INVALID,
					_PT("Content of 'duration' is not an unsigned integer greater zero."));
			return false;
		} else {
			this->track->setDuration(duration);
		}
		break;
/*
	case TAG_PLAYLIST_TRACKLIST_TRACK_EXTENSION:
		break;
*/
	case TAG_PLAYLIST_TRACKLIST_TRACK_IDENTIFIER:
		if (!isURI(finalAccum)) {
			setError(SPIFF_PARSER_ERROR_CONTENT_INVALID,
					_PT("Content of 'identifier' is not a URI."));
			return false;
		} else {
			this->track->addIdentifier(finalAccum);
		}
		break;

	case TAG_PLAYLIST_TRACKLIST_TRACK_IMAGE:
		if (!isURI(finalAccum)) {
			setError(SPIFF_PARSER_ERROR_CONTENT_INVALID,
					_PT("Content of 'image' is not a URI."));
			return false;
		} else {
			this->track->setImage(finalAccum);
		}
		break;

	case TAG_PLAYLIST_TRACKLIST_TRACK_INFO:
		if (!isURI(finalAccum)) {
			setError(SPIFF_PARSER_ERROR_CONTENT_INVALID,
					_PT("Content of 'info' is not a URI."));
			return false;
		} else {
			this->track->setInfo(finalAccum);
		}
		break;

	case TAG_PLAYLIST_TRACKLIST_TRACK_LINK:
		if (!isURI(finalAccum)) {
			setError(SPIFF_PARSER_ERROR_CONTENT_INVALID,
					_PT("Content of 'link' is not a URI."));
			return false;
		} else {
			this->track->addLink(this->lastRelValue.c_str(), finalAccum);
		}
		break;

	case TAG_PLAYLIST_TRACKLIST_TRACK_LOCATION:
		if (!isURI(finalAccum)) {
			setError(SPIFF_PARSER_ERROR_CONTENT_INVALID,
					_PT("Content of 'location' is not a URI."));
			return false;
		} else {
			this->track->addLocation(finalAccum);
		}
		break;

	case TAG_PLAYLIST_TRACKLIST_TRACK_META:
		this->track->addMeta(this->lastRelValue.c_str(), finalAccum);
		break;

	case TAG_PLAYLIST_TRACKLIST_TRACK_TITLE:
		this->track->setTitle(finalAccum);
		break;

	case TAG_PLAYLIST_TRACKLIST_TRACK_TRACKNUM:
		int trackNum;
		if (!extractInteger(finalAccum, 1, &trackNum)) {
			setError(SPIFF_PARSER_ERROR_CONTENT_INVALID,
					_PT("Content of 'trackNum' is not an unsigned integer."));
			return false;
		} else {
			this->track->setTrackNum(trackNum);
		}
		break;

	}

	this->accum.clear();
	return true;
}



void SpiffReader::handleEnd(const XML_Char * name) {
	bool extensionLeft = false;
	int pushBackTag;
	if (this->insideExtension) {
		switch (this->stack.getSize()) {
		case 2:
			if (this->stack.top() == TAG_PLAYLIST_EXTENSION) {
				pushBackTag = TAG_PLAYLIST_EXTENSION;
				extensionLeft = true;
			}
			break;

		case 4:
			if (this->stack.top() == TAG_PLAYLIST_TRACKLIST_TRACK_EXTENSION) {
				pushBackTag = TAG_PLAYLIST_TRACKLIST_TRACK_EXTENSION;
				extensionLeft = true;
			}
			break;

		}

		if (!handleExtensionEnd(name)) {
			stop();
			return;
		}

		if (extensionLeft) {
			this->insideExtension = false;
			this->stack.push(pushBackTag);
		} else {
			return;
		}
	}

	bool res = false;
	switch (this->stack.getSize()) {
	case 1:
		res = handleEndOne(name);
		break;

	case 2:
		res = handleEndTwo(name);
		break;

	case 3:
		res = handleEndThree(name);
		break;

	case 4:
		res = handleEndFour(name);
		break;

	}

	if (!res) {
		stop();
		return;
	}

	// Prevent popping twice
//	if (!extensionLeft) {
		this->stack.pop();
//	}
}



void SpiffReader::handleCharacters(const XML_Char * s, int len) {
	if (this->insideExtension) {
		if (!handleExtensionCharacters(s, len)) {
			stop();
		}
		return;
	}

	const int LEVEL = this->stack.getSize();
	const int TAG = this->stack.top();

	switch (this->stack.getSize()) {
	case 2:
		switch (this->stack.top()) {
		case TAG_PLAYLIST_TRACKLIST:
		case TAG_PLAYLIST_ATTRIBUTION:
			break;

		default:
			this->accum.append(s, len);

		}
		break;

	case 3:
		switch (this->stack.top()) {
		case TAG_PLAYLIST_ATTRIBUTION_IDENTIFIER:
		case TAG_PLAYLIST_ATTRIBUTION_LOCATION:
			this->accum.append(s, len);
			break;

		}
		break;

	case 4:
		this->accum.append(s, len);
		break;

	}
}



void SpiffReader::stop() {
	// Remove handlers
	XML_SetElementHandler(this->parser, NULL, NULL);
	XML_SetCharacterDataHandler(this->parser, NULL);

	// Full stop
	XML_StopParser(this->parser, XML_FALSE);
}



bool SpiffReader::handlePlaylistAttribs(const XML_Char ** atts) {
	bool versionFound = false;
	bool xmlnsFound = false;

	while (atts[0] != NULL) {
		if (!::PORT_STRCMP(atts[0], _PT("version"))) {
			int dummyVersion;
			if (!extractInteger(atts[1], 0, &dummyVersion) || (dummyVersion > 1)) {
				setError(SPIFF_PARSER_ERROR_ATTRIBUTE_INVALID,
						_PT("Version must be '0' or '1', not '%s'."),
						atts[1]);
				return false;
			}
			this->version = dummyVersion;
			versionFound = true;
		} else if (!::PORT_STRCMP(atts[0], _PT("xmlns"))) {
			if (::PORT_STRCMP(atts[1], _PT("http://xspf.org/ns/0/"))) {
				setError(SPIFF_PARSER_ERROR_ATTRIBUTE_INVALID,
						_PT("Namespace must be 'http://xspf.org/ns/0/', not '%s'."),
						atts[1]);
				return false;
			}
			xmlnsFound = true;
		} else {
			setError(SPIFF_PARSER_ERROR_ATTRIBUTE_FORBIDDEN, _PT("Attribute '%s' not allowed."), atts[0]);
			return false;
		}
		atts += 2;
	}

	if (!versionFound) {
		setError(SPIFF_PARSER_ERROR_ATTRIBUTE_MISSING, _PT("Attribute 'version' missing."));
		return false;
	}

	if (!xmlnsFound) {
		setError(SPIFF_PARSER_ERROR_ATTRIBUTE_MISSING, _PT("Attribute 'xmlns' missing."));
		return false;
	}

	return true;
}



bool SpiffReader::handleMetaLinkAttribs(const XML_Char ** atts) {
	if (atts[0] == NULL) {
		setError(SPIFF_PARSER_ERROR_ATTRIBUTE_MISSING, _PT("Attribute 'rel' missing."));
		return false;
	}

	if (atts[2] != NULL) {
		setError(SPIFF_PARSER_ERROR_ATTRIBUTE_FORBIDDEN, _PT("Attribute '%s' not allowed."), atts[2]);
		return false;
	}

	if (::PORT_STRCMP(atts[0], _PT("rel"))) {
		setError(SPIFF_PARSER_ERROR_ATTRIBUTE_FORBIDDEN, _PT("Attribute '%s' not allowed."), atts[0]);
		return false;
	}

	if (!isURI(atts[1])) {
		setError(SPIFF_PARSER_ERROR_ATTRIBUTE_INVALID, _PT("Attribute 'rel' is not a URI."));
		return false;
	}

	return true;
}



bool SpiffReader::handleExtensionAttribs(const XML_Char ** atts) {
	if (atts[0] == NULL) {
		setError(SPIFF_PARSER_ERROR_ATTRIBUTE_MISSING, _PT("Attribute 'application' missing."));
		return false;
	}

	if (atts[2] != NULL) {
		setError(SPIFF_PARSER_ERROR_ATTRIBUTE_FORBIDDEN, _PT("Attribute '%s' not allowed."), atts[2]);
		return false;
		
	}

	if (::PORT_STRCMP(atts[0], _PT("application"))) {
		setError(SPIFF_PARSER_ERROR_ATTRIBUTE_FORBIDDEN, _PT("Attribute '%s' not allowed."), atts[0]);
		return false;
	}

	if (!isURI(atts[1])) {
		setError(SPIFF_PARSER_ERROR_ATTRIBUTE_INVALID, _PT("Attribute 'application' is not a URI."));
		return false;
	}

	return true;
}



bool SpiffReader::handleNoAttribs(const XML_Char ** atts) {
	if (atts[0] != NULL) {
		setError(SPIFF_PARSER_ERROR_ATTRIBUTE_FORBIDDEN, _PT("Attribute '%s' not allowed."), atts[0]);
		return false;
		
	}

	return true;
}



SpiffTrack * SpiffReader::newTrack() {
	return new SpiffTrack();
}



SpiffProps * SpiffReader::newProps() {
	return new SpiffProps();
}



void SpiffReader::setError(int code, const XML_Char * text, const XML_Char * param) {
	this->errorCode = code;
	if (param == NULL) {
		this->errorText.assign(text);
	} else {
		XML_Char * const final = new XML_Char[::PORT_STRLEN(text) + ::PORT_STRLEN(param) + 1];
		::PORT_SPRINTF(final, text, param);
		this->errorText.assign(final);
		delete [] final;
	}
	this->errorLine = XML_GetCurrentLineNumber(this->parser);
}



void SpiffReader::clearError() {
	this->errorCode = SPIFF_PARSER_SUCCESS;
	this->errorLine = -1;
	this->errorText.clear();
}



const XML_Char * SpiffReader::getErrorText() {
	return this->errorText.c_str();
}



int SpiffReader::getErrorLine() {
	return this->errorLine;
}



/*static*/ inline void SpiffReader::masterStart(void * userData, const XML_Char * name, const XML_Char ** atts) {
	SpiffReader * const parser = reinterpret_cast<SpiffReader *>(userData);
	parser->handleStart(name, atts);
}



/*static*/ inline void SpiffReader::masterEnd(void * userData, const XML_Char * name) {
	SpiffReader * const parser = reinterpret_cast<SpiffReader *>(userData);
	parser->handleEnd(name);
}



/*static*/ inline void SpiffReader::masterCharacters(void * userData, const XML_Char * s, int len) {
	SpiffReader * const parser = reinterpret_cast<SpiffReader *>(userData);
	parser->handleCharacters(s, len);
}



/*static*/ bool SpiffReader::isURI(const XML_Char * text) {
	// http://en.wikipedia.org/wiki/URI_scheme

	// <scheme name>
	int counter = 0;
	for (;;) {
		switch (*text) {
		case _PT('\0'):
			return false;

		case _PT(':'):
            return (counter != 0);

		case _PT('+'):
		case _PT('-'):
		case _PT('.'):
			counter++;
			text++;
			break;

		default:
			if (((*text >= _PT('a')) && (*text <= _PT('z')))
					|| ((*text >= _PT('A')) && (*text <= _PT('Z')))
					|| ((*text >= _PT('0')) && (*text <= _PT('9')))) {
				counter++;
				text++;
			} else {
				return false;
			}
			break;

		}
	}
}



/*static*/ bool SpiffReader::extractInteger(const XML_Char * text, int inclusiveMinimum, int * output) {
	int & number = *output;
	number = ::PORT_ATOI(text);
	if ((number < inclusiveMinimum)
			|| ((number == 0) && ::PORT_STRCMP(text, _PT("0")))) {
		return false;
	} else {
		return true;
	}
}



/*static*/ bool SpiffReader::extractDateTime(const XML_Char * text, SpiffDateTime * output) {
	// http://www.w3.org/TR/xmlschema-2/#dateTime-lexical-representation
	// '-'? yyyy '-' mm '-' dd 'T' hh ':' mm ':' ss ('.' s+)? (zzzzzz)?
	const XML_Char * const textBackup = text;

	// '-'?
	const bool leadingMinus = (*text == _PT('-'));
	if (leadingMinus) {
		text++;
	}

	// yyyy
	if ((::PORT_STRNCMP(text, _PT("0001"), 4) < 0) || (::PORT_STRNCMP(_PT("9999"), text, 4) < 0)) {
		return false;
	}
	output->year = Toolbox::PORT_ANTOI(text, 4);
	text += 4;

	// '-' mm
	if ((::PORT_STRNCMP(text, _PT("-01"), 3) < 0) || (::PORT_STRNCMP(_PT("-12"), text, 3) < 0)) {
		return false;
	}
	output->month = Toolbox::PORT_ANTOI(text + 1, 2);
	text += 3;

	// '-' dd
	if ((::PORT_STRNCMP(text, _PT("-01"), 3) < 0) || (::PORT_STRNCMP(_PT("-31"), text, 3) < 0)) {
		return false;
	}
	output->day = Toolbox::PORT_ANTOI(text + 1, 2);
	text += 3;

	// Month specific day check
	switch (output->month) {
	case 2:
		switch (output->day) {
		case 31:
		case 30:
			return false;

		case 29:
			if (((output->year % 400) != 0) && (((output->year % 4) != 0)
					|| ((output->year % 100) == 0))) {
				// Not a leap year
				return false;
			}
			break;

		case 28:
			break;
		}
		break;

	case 4:
	case 6:
	case 9:
	case 11:
		if (output->day > 30) {
			return false;
		}
		break;

	}

	// 'T' hh
	if ((::PORT_STRNCMP(text, _PT("T00"), 3) < 0) || (::PORT_STRNCMP(_PT("T23"), text, 3) < 0)) {
		return false;
	}
	output->hour = Toolbox::PORT_ANTOI(text + 1, 2);
	text += 3;

	// ':' mm
	if ((::PORT_STRNCMP(text, _PT(":00"), 3) < 0) || (::PORT_STRNCMP(_PT(":59"), text, 3) < 0)) {
		return false;
	}
	output->minutes = Toolbox::PORT_ANTOI(text + 1, 2);
	text += 3;

	// ':' ss
	if ((::PORT_STRNCMP(text, _PT(":00"), 2) < 0) || (::PORT_STRNCMP(_PT(":59"), text, 2) < 0)) {
		return false;
	}
	output->seconds = Toolbox::PORT_ANTOI(text + 1, 2);
	text += 3;

	// ('.' s+)?
	if (*text == _PT('.')) {
		text++;
		int counter = 0;
		while ((*text >= _PT('0')) && (*text <= _PT('9'))) {
			text++;
			counter++;
		}
		if (counter == 0) {
			return false;
		} else {
			// The fractional second string, if present, must not end in '0'
			if (*(text - 1) == _PT('0')) {
				return false;
			}
		}
	}

	// (zzzzzz)? := (('+' | '-') hh ':' mm) | 'Z'
	const XML_Char * const timeZoneStart = text;
	switch (*text) {
	case _PT('+'):
	case _PT('-'):
		{
			text++;
			if ((::PORT_STRNCMP(text, _PT("00"), 2) < 0) || (::PORT_STRNCMP(_PT("14"), text, 2) < 0)) {
				return false;
			}
			output->distHours = Toolbox::PORT_ANTOI(text, 2);
			text += 2;

			if ((::PORT_STRNCMP(text, _PT(":00"), 3) < 0) || (::PORT_STRNCMP(_PT(":59"), text, 3) < 0)) {
				return false;
			}
			output->distMinutes = Toolbox::PORT_ANTOI(text + 1, 2);
			if ((output->distHours == 14) && (output->distMinutes != 0)) {
				return false;
			}
			text += 3;

			if (*text != _PT('\0')) {
				return false;
			}

			if (*timeZoneStart == _PT('-')) {
				output->distHours = -output->distHours;
				output->distMinutes = -output->distMinutes;
			}
		}
		break;

	case _PT('Z'):
		text++;
		if (*text != _PT('\0')) {
			return false;
		}
		// NO BREAK

	case _PT('\0'):
		output->distHours = 0;
		output->distMinutes = 0;
		break;

	default:
		return false;

	}

	return true;
}



}
