/**
 * libSpiff - XSPF playlist handling library
 *
 * Copyright (C) 2006 Sebastian Pipping
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Sebastian Pipping, webmaster@hartwork.org
 */

#ifndef SPIFF_READER_H
#define SPIFF_READER_H 1



#include "SpiffDefines.h"
#include "SpiffStack.h"
#include <string>
using namespace std;

namespace Spiff {



/*
playlist									1
	title							?
	creator							?
	annotation						?
	info							?
	location						?
	identifier						?
	image							?
	date							?
	license							?
	attribution						?
		location						*
		identifier						*
	link								*
	meta								*
	extension							*
		...								*
	trackList								1
		track							+/*
			location					*
			identifier					*
			title					?
			creator					?
			annotation				?
			info					?
			image					?
			album 					?
			trackNum (uint > 0)		?
			duration (uint)			?
			link						*
			meta						*
			extension					*
*/



enum SpiffParseReturnCode {

	SPIFF_PARSER_SUCCESS,

	SPIFF_PARSER_ERROR_NO_INPUT,

	SPIFF_PARSER_ERROR_ELEMENT_TOOMANY,
	SPIFF_PARSER_ERROR_ELEMENT_FORBIDDEN,
	SPIFF_PARSER_ERROR_ELEMENT_MISSING,

	SPIFF_PARSER_ERROR_ATTRIBUTE_INVALID,
	SPIFF_PARSER_ERROR_ATTRIBUTE_MISSING,
	SPIFF_PARSER_ERROR_ATTRIBUTE_FORBIDDEN,

	SPIFF_PARSER_ERROR_CONTENT_INVALID,

	SPIFF_PARSER_ERROR_EXPAT // This one must come last!

};



class SpiffProps;
class SpiffDateTime;
class SpiffTrack;
class SpiffReaderCallback;



class SpiffReader {

protected:
	SpiffStack stack;
	SpiffProps * props;
	SpiffTrack * track;
	int version;

private:
	XML_Parser parser;
	SpiffReaderCallback * callback;
	basic_string<XML_Char> accum;
	basic_string<XML_Char> lastRelValue;

	int errorCode;
	int errorLine;
	basic_string<XML_Char> errorText;

	bool insideExtension;

	bool firstPlaylistAnnotation;
	bool firstPlaylistAttribution;
	bool firstPlaylistCreator;
	bool firstPlaylistDate;
	bool firstPlaylistIdentifier;
	bool firstPlaylistImage;
	bool firstPlaylistInfo;
	bool firstPlaylistLicense;
	bool firstPlaylistLocation;
	bool firstPlaylistTitle;
	bool firstPlaylistTrackList;

	bool firstTrackTitle;
	bool firstTrackCreator;
	bool firstTrackAnnotation;
	bool firstTrackInfo;
	bool firstTrackImage;
	bool firstTrackAlbum;
	bool firstTrackTrackNum;
	bool firstTrackDuration;

	bool firstTrack;

public:
	SpiffReader();
	~SpiffReader();

	int parse(const XML_Char * filename, SpiffReaderCallback * callback);
	const XML_Char * getErrorText();
	int getErrorLine();

protected:
	virtual bool handleExtensionStart(const XML_Char * name, const XML_Char ** atts);
	virtual bool handleExtensionEnd(const XML_Char * name);
	virtual bool handleExtensionCharacters(const XML_Char * s, int len);

	virtual SpiffTrack * newTrack();
	virtual SpiffProps * newProps();

	virtual void makeReusable();

	void setError(int code, const XML_Char * text, const XML_Char * param = NULL);
	void stop();

private:
	void handleStart(const XML_Char * name, const XML_Char ** atts);
	bool handleStartOne(const XML_Char * name, const XML_Char ** atts);
	bool handleStartTwo(const XML_Char * name, const XML_Char ** atts);
	bool handleStartThree(const XML_Char * name, const XML_Char ** atts);
	bool handleStartFour(const XML_Char * name, const XML_Char ** atts);

	void handleCharacters(const XML_Char * s, int len);

	void handleEnd(const XML_Char * name);
	bool handleEndOne(const XML_Char * name);
	bool handleEndTwo(const XML_Char * name);
	bool handleEndThree(const XML_Char * name);
	bool handleEndFour(const XML_Char * name);

	bool handlePlaylistAttribs(const XML_Char ** atts);
	bool handleMetaLinkAttribs(const XML_Char ** atts);
	bool handleExtensionAttribs(const XML_Char ** atts);
	bool handleNoAttribs(const XML_Char ** atts);

	void clearError();

public:
	static bool isURI(const XML_Char * text);
	static bool extractInteger(const XML_Char * text, int inclusiveMinimum, int * output);
	static bool extractDateTime(const XML_Char * text, SpiffDateTime * output);

private:
	static void masterStart(void * userData, const XML_Char * name, const XML_Char ** atts);
	static void masterEnd(void * userData, const XML_Char * name);
	static void masterCharacters(void * userData, const XML_Char * s, int len);

};



}

#endif // SPIFF_READER_H
