/**
 * libSpiff - XSPF playlist handling library
 *
 * Copyright (C) 2006 Sebastian Pipping
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Sebastian Pipping, webmaster@hartwork.org
 */

#include "SpiffTrack.h"
#include "SpiffToolbox.h"
#include <string.h>

namespace Spiff {



SpiffTrack::SpiffTrack() : album(NULL), image(NULL), info(NULL), annotation(NULL),
		creator(NULL), title(NULL), trackNum(-1), duration(-1), locations(NULL),
		identifiers(NULL), links(NULL), metas(NULL) {
	
}



SpiffTrack::~SpiffTrack() {
	if (this->album != NULL) {
		delete [] this->album;
	}
	if (this->image != NULL) {
		delete [] this->image;
	}
	if (this->info != NULL) {
		delete [] this->info;
	}
	if (this->annotation != NULL) {
		delete [] this->annotation;
	}
	if (this->creator != NULL) {
		delete [] this->creator;
	}
	if (this->title != NULL) {
		delete [] this->title;
	}
	if (this->locations != NULL) {
		SpiffStringList::const_iterator iter = this->locations->begin();
		while (iter != this->locations->end()) {
			delete [] *iter;
			iter++;
		}
		delete this->locations;
	}
	if (this->identifiers != NULL) {
		SpiffStringList::const_iterator iter = this->identifiers->begin();
		while (iter != this->identifiers->end()) {
			delete [] *iter;
			iter++;
		}
		delete this->identifiers;
	}
	if (this->links != NULL) {
		SpiffMetaLinkList::const_iterator iter = this->links->begin();
		while (iter != this->links->end()) {
			const SpiffMetaLink & entry = *iter;
			delete [] entry.first;
			delete [] entry.second;
			iter++;
		}
		delete this->links;
	}
	if (this->metas != NULL) {
		SpiffMetaLinkList::const_iterator iter = this->metas->begin();
		while (iter != this->metas->end()) {
			const SpiffMetaLink & entry = *iter;
			delete [] entry.first;
			delete [] entry.second;
			iter++;
		}
		delete this->metas;
	}
}



void SpiffTrack::setAnnotation(const XML_Char * annotation) {
	Toolbox::deleteNewAndCopy(&this->annotation, annotation);
}



void SpiffTrack::setCreator(const XML_Char * creator) {
	Toolbox::deleteNewAndCopy(&this->creator, creator);
}



void SpiffTrack::setInfo(const XML_Char * info) {
	Toolbox::deleteNewAndCopy(&this->info, info);
}



void SpiffTrack::setImage(const XML_Char * image) {
	Toolbox::deleteNewAndCopy(&this->image, image);
}



void SpiffTrack::setAlbum (const XML_Char * album) {
	Toolbox::deleteNewAndCopy(&this->album , album);
}



void SpiffTrack::setTitle(const XML_Char * title) {
	Toolbox::deleteNewAndCopy(&this->title, title);
}



void SpiffTrack::setTrackNum(int trackNum) {
	this->trackNum = trackNum;
}



void SpiffTrack::setDuration(int duration) {
	this->duration = duration;
}



int SpiffTrack::getTrackNum() {
	return this->trackNum;
}

int SpiffTrack::getDuration() {
	return this->duration;
}



XML_Char * SpiffTrack::stealTitle() {
	return Toolbox::getSetNull<XML_Char>(&this->title);
}



XML_Char * SpiffTrack::stealAnnotation() {
	return Toolbox::getSetNull<XML_Char>(&this->annotation);
}



XML_Char * SpiffTrack::stealCreator() {
	return Toolbox::getSetNull<XML_Char>(&this->creator);
}



XML_Char * SpiffTrack::stealInfo() {
	return Toolbox::getSetNull<XML_Char>(&this->info);
}



XML_Char * SpiffTrack::stealImage() {
	return Toolbox::getSetNull<XML_Char>(&this->image);
}



XML_Char * SpiffTrack::stealAlbum() {
	return Toolbox::getSetNull<XML_Char>(&this->album);
}



void SpiffTrack::addIdentifier(const XML_Char * identifier) {
	if (this->identifiers == NULL) {
		this->identifiers = new SpiffStringList();
	}
	this->identifiers->push_back(Toolbox::newAndCopy(identifier));
}



void SpiffTrack::addLocation(const XML_Char * location) {
	if (this->locations == NULL) {
		this->locations = new SpiffStringList();
	}
	this->locations->push_back(Toolbox::newAndCopy(location));
}



void SpiffTrack::addLink(const XML_Char * rel, const XML_Char * content) {
	if (this->links == NULL) {
		this->links = new SpiffMetaLinkList();
	}
	pair<XML_Char *, XML_Char *> entry(Toolbox::newAndCopy(rel), Toolbox::newAndCopy(content));
	this->links->push_back(entry);
}



void SpiffTrack::addMeta(const XML_Char * rel, const XML_Char * content) {
	if (this->metas == NULL) {
		this->metas = new SpiffMetaLinkList();
	}
	pair<XML_Char *, XML_Char *> entry(Toolbox::newAndCopy(rel), Toolbox::newAndCopy(content));
	this->metas->push_back(entry);
}



SpiffStringList * SpiffTrack::stealIdentifier() {
	return Toolbox::getSetNull<SpiffStringList>(&this->identifiers);
}



SpiffStringList * SpiffTrack::stealLocation() {
	return Toolbox::getSetNull<SpiffStringList>(&this->locations);
}



SpiffMetaLinkList * SpiffTrack::stealLink() {
	return Toolbox::getSetNull<SpiffMetaLinkList>(&this->links);
}



SpiffMetaLinkList * SpiffTrack::stealMeta() {
	return Toolbox::getSetNull<SpiffMetaLinkList>(&this->metas);
}



const XML_Char * SpiffTrack::getAlbum() {
	return this->album;
}


const XML_Char * SpiffTrack::getImage() {
	return this->image;
}


const XML_Char * SpiffTrack::getInfo() {
	return this->info;
}


const XML_Char * SpiffTrack::getAnnotation() {
	return this->annotation;
}


const XML_Char * SpiffTrack::getCreator() {
	return this->creator;
}


const XML_Char * SpiffTrack::getTitle() {
	return this->title;
}



const SpiffStringList * SpiffTrack::getLocations() {
	return this->locations;
}


const SpiffStringList * SpiffTrack::getIdentifiers() {
	return this->identifiers;
}



const SpiffMetaLinkList * SpiffTrack::getLinks() {
	return this->links;
}


const SpiffMetaLinkList * SpiffTrack::getMetas() {
	return this->metas;
}



}
