/**
 * libSpiff - XSPF playlist handling library
 *
 * Copyright (C) 2006 Sebastian Pipping
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Sebastian Pipping, webmaster@hartwork.org
 */

#include "SpiffWriter.h"
#include "SpiffAbstractTrackWriter.h"
#include "SpiffAbstractPropsWriter.h"

namespace Spiff {



SpiffWriter::SpiffWriter(int version) : trackList(new basic_ostringstream<XML_Char>()),
		trackCount(0), version(version) {
	// Fix invalid version
	if ((this->version < 0) || (this->version > 1)) {
		this->version = 1;
	}
}



SpiffWriter::~SpiffWriter() {
	if (this->trackList != NULL) {
		delete this->trackList;
	}
}



void SpiffWriter::addTrack(SpiffAbstractTrackWriter & trackWriter) {
	trackWriter.init(*this->trackList, this->version);
	trackWriter.write();
	this->trackCount++;
}



int SpiffWriter::write(const XML_Char * filename,
		SpiffAbstractPropsWriter & propsWriter) {
	// Note: trackList buffer is only cleared on success
	basic_ostringstream<XML_Char> finalStream;

	// Open file
	FILE * const file = ::PORT_FOPEN(filename, _PT("wb"));
	if (file == NULL) {
		return SPIFF_WRITER_ERROR;
	}

	// Head
	propsWriter.init(finalStream, this->version, this->trackCount);
	propsWriter.writeHead();

	// Body
	basic_string<XML_Char> finalTrackList = this->trackList->str();
	const XML_Char * const rawFinalTrackList = finalTrackList.c_str();
	const int rawFinalTrackListLen = static_cast<int>(::PORT_STRLEN(rawFinalTrackList));
	finalStream.write(rawFinalTrackList, rawFinalTrackListLen);

	// Foot
	propsWriter.writeFoot();

	// Write to file
	basic_string<XML_Char> final = finalStream.str();
	const XML_Char * const rawFinal = final.c_str();
	const int rawFinalLen = static_cast<int>(::PORT_STRLEN(rawFinal));

	// UTF-8 conversion on Unicode Windows
#if (defined(UNICODE) && (defined(__WIN32__) || defined(WIN32)))
	char * rawFinalUtf8 = new char[rawFinalLen * 4];
	const int rawFinalUtf8Len = ::WideCharToMultiByte(CP_UTF8, 0, rawFinal, rawFinalLen, rawFinalUtf8, rawFinalLen * 4, NULL, NULL);
	::fwrite(rawFinalUtf8, 1, rawFinalUtf8Len, file);
	delete [] rawFinalUtf8;
#else
	::fwrite(rawFinal, sizeof(XML_Char), rawFinalLen, file);
#endif
	::fclose(file);

	// Make instance reusable
	delete this->trackList;
	this->trackList = new basic_ostringstream<XML_Char>;
	this->trackCount = 0;

	return SPIFF_WRITER_SUCCESS;
}



}
